/*
 * Xournal++
 *
 * Current git repo info
 *
 * @author Xournal++ Team
 * https://github.com/xournalpp/xournalpp
 *
 * @license GNU GPLv2 or later
 */

#pragma once


/* --- Current git repo info --- */

/**
 * Origin URL of current repository
 */
constexpr auto GIT_ORIGIN_URL = "/srcdest/xournalpp";

/**
 * Git origin repo owner
 */
constexpr auto GIT_ORIGIN_OWNER = "srcdest";

/**
 * Git origin repo name
 */
constexpr auto GIT_ORIGIN_REPO = "xournalpp";

/**
 * Current branch of git repo
 */
constexpr auto GIT_BRANCH = "makepkg";

/**
 * Current git commit id (with "-dirty" appended if dirty source tree)
 */
constexpr auto GIT_COMMIT_ID = "922f80df";
